class IncorrectParameters(Exception):
    '''Raised when the incorrect configuration of parameters is passed into a Class'''
    pass

class ParseResultsError(RuntimeWarning):
    '''Raised when innvocation returned an error from a remote machine/comamnd'''
    pass


from pypsrp.client import Client
from pypsrp.powershell import PowerShell, RunspacePool
from pypsrp.wsman import WSMan
from pypsrp.shell import WinRS 

from ..core import Core
from .state import State
from .parseresults import ParseResultsState


class InnvocationState(State, Core):
    """
    The state which indicates the invocation of a command
    """

    __win_client = None

    def __handle_windows_errors(self, stream):
        return_list = []
        for item in stream.error:
            return_list.append({
                'type': 'error',
                'value': str(item)
            })
        for item in stream.debug:
            return_list.append({
                'type': 'debug',
                'value': str(item)
            })
        for item in stream.information:
            return_list.append({
                'type': 'information',
                'value': str(item)
            })
        for item in stream.verbose:
            return_list.append({
                'type': 'verbose',
                'value': str(item)
            })
        for item in stream.warning:
            return_list.append({
                'type': 'warning',
                'value': str(item)
            })
        return return_list

    def __create_win_client(self, hostinfo):
        self.__win_client = Client(
            hostinfo.hostname,
            username=hostinfo.username,
            password=hostinfo.password,
            ssl=hostinfo.verify_ssl
        )

    def __invoke_cmd(self, command):
        if not self.__win_client:
            self.__create_win_client(self.hostinfo)
        stdout, stderr, rc = self.__win_client.execute_cmd(command)
        # NOTE: rc (return code of process) should equal 0 but we are not adding logic here this is handled int he ParseResultsState class
        if stderr:
            self.__logger.error('{host} responded with the following message(s): {message}'.format(
                host=self.hostinfo.hostname,
                message=stderr
            ))
        return ParseResultsState(stdout)

    def __invoke_powershell(self, command):
        if not self.__win_client:
            self.__create_win_client(self.hostinfo)
        output, streams, had_errors = self.__win_client.execute_ps(command)
        if not output:
            output = self.__handle_windows_errors(streams)
        if had_errors:
            self.__logger.error('{host} responded with the following message(s): {message}'.format(
                host=self.hostinfo.hostname,
                message=self.__handle_windows_errors(streams)
            ))
        return ParseResultsState(output)

    def __invoke_ssh(self,command):
        import paramiko
        ssh = paramiko.SSHClient()
        ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        if self.hostinfo.ssh_key_path:
            ssh.connect(
                self.hostinfo.hostname,
                port=self.hostinfo.port,
                username=self.hostinfo.username,
                key_filename=self.hostinfo.ssh_key_path
            )
        elif self.hostinfo.private_key_string:
            ssh.connect(
                self.hostinfo.hostname,
                port=self.hostinfo.port,
                username=self.hostinfo.username,
                pkey=self.hostinfo.private_key_string
            )
        elif self.hostinfo.password:
            ssh.connect(
                self.hostinfo.hostname,
                port=self.hostinfo.port,
                username=self.hostinfo.username,
                password=self.hostinfo.password,
                timeout=self.hostinfo.timeout
            )
        else:
            raise AttributeError('Please provide either a ssh_key_path or a password')
        out = None
        ssh_stdin, ssh_stdout, ssh_stderr = ssh.exec_command(command)
        out = ssh_stdout.read()
        ssh_stdin.flush()
        ssh.close()
        return ParseResultsState(out)

    def invoke(self, hostinfo, command_type, command):
        self.hostinfo = hostinfo
        if command_type == 'powershell':
            result = self.__invoke_powershell(command)
        elif command_type == 'cmd':
            result = self.__invoke_cmd(command)
        elif command_type == 'ssh':
            result = self.__invoke_ssh(command)
        return result



import os
import yaml

from .utils.logger import LoggingBase


class Core(metaclass=LoggingBase):

    def get_abs_path(self, value):
        """Formats and returns the absolute path for a path value

        Args:
            value (str): A path string in many different accepted formats

        Returns:
            str: The absolute path of the provided string
        """
        return os.path.abspath(os.path.expanduser(os.path.expandvars(value)))

    def parse_config_file(self, config_file_path):
        config = None
        return_list = []
        try:
            with open(self.get_abs_path(config_file_path), 'r', encoding="utf-8") as f:
                config = yaml.load(f.read(), Loader=yaml.SafeLoader)
        except:
            # windows does not like get_abs_path so casting to string
            with open(str(config_file_path), 'r', encoding="utf-8") as f:
                config = yaml.load(f.read(), Loader=yaml.SafeLoader)
        if config and config.get('inventory') and config.get('run_conditions'):
            from .runner import Runner, Host
            for key,val in config['run_conditions'].items():
                for inventory in val['inventories']:
                    if config['inventory'].get(inventory):
                        host_list = []
                        inputs = config['inventory'][inventory]['inputs']
                        for host in config['inventory'][inventory]['hosts']:
                            host_list.append(
                                Host(
                                    hostname=host,
                                    username=inputs.get('username'),
                                    password=inputs.get('password'),
                                    verify_ssl=inputs['verify_ssl'] if inputs.get('verify_ssl') else False,
                                    ssh_key_path=inputs.get('ssh_key_path'),
                                    private_key_string=inputs.get('private_key_string'),
                                    port=inputs['port'] if inputs.get('port') else 22,
                                    timeout=inputs['timeout'] if inputs.get('timeout') else 5
                                )
                            )
                        if host_list:
                            return_list.append(
                                Runner(
                                    executor=val['executor'],
                                    command=val['command'],
                                    hosts=host_list
                                )
                            )
        return return_list



from .core import Core
from .runner import Runner, Host
from .states.creation import CreationState


class Rudder(Core):
    """ 
    A simple state machine that runs commands remotely on the targeted system

        ## Create Host Object and running a command
            from rudder import Host

            my_host_list = []
            my_host_list.append(
                Host(
                    hostname='10.10.32.100',
                    username='admin',
                    password='secret_password1'
                )
            )

            rudder = Rudder(hosts=my_host_list)
            for result in rudder.execute(executor='powershell', command='Get-ChildItem -Path "C:\" -Recurse'):
                print(result)

        ## Providing a config_file_path
            from rudder import Rudder

            rudder = Rudder(config_file_path='~/path_to_config_file.yml')
            for result in rudder.execute():
                print(result)
    """

    def execute(self, 
        host: Host = [None],
        config_file_path: str  = None,
        executor: str = None, 
        command: str = None):
        """The execute method is the main entry point to execute the provided command on a remote host.

        You must create a host object and pass it to this method along with the executor and command to execute.

        Args:
            host (Host, optional): Create a list of one or more Host objects and provide this as input. Defaults to [None].
            config_file_path (str, optional): A path to a local config file. Defaults to None.
            executor (str, optional): A executor of powershell or cmd or ssh is needed. 
                                      Default is None but must be provided if not using a config_file_path
            command (str, optional): The command to run on a specified remote host
                                     Default is None but must be provided if not using a config_file_path

        Returns:
            str: Returns the results from a remote command
        """
        if host:
            self.__host_info = host
        else:
            self.__host_info = None
        if config_file_path:
            self.__configs = self.parse_config_file(config_file_path)
        else:
            self.__configs = None
        self.__logger.debug('Intializing default state of CreationState')
        self.state = CreationState()
        if not self.__configs and self.__host_info and executor in ['powershell', 'cmd', 'ssh'] and command:
            self.__configs = [Runner(
                executor=executor,
                command=command,
                hosts=self.__host_info
            )]
        elif self.__configs:
            pass
        else:
            raise ValueError('You must provide a valid executor and command or a properly formatted config file')
        if self.__configs:
            for config in self.__configs:
                for host in config.hosts:
                    finished = False
                    while not finished:
                        if str(self.state) == 'CreationState':
                            self.__logger.debug('Running CreationState on_event')
                            self.state = self.state.on_event(config.executor, config.command)
                        if str(self.state) == 'InnvocationState':
                            self.__logger.debug('Running InnvocationState on_event')
                            self.state = self.state.invoke(host, config.executor, config.command)
                        if str(self.state) == 'ParseResultsState':
                            self.__logger.debug('Running ParseResultsState on_event')
                            yield self.state.on_event()
                            finished = True



import typing
import attr

from .core import Core


@attr.s
class Host:

    hostname = attr.ib(type=str)
    username = attr.ib(default=None, type=str)
    password = attr.ib(default=None, type=str)
    verify_ssl = attr.ib(default=False, type=bool)
    ssh_key_path = attr.ib(default=None, type=str)
    private_key_string = attr.ib(default=None, type=str)
    port = attr.ib(default=22, type=int)
    timeout = attr.ib(default=5, type=int)

    @ssh_key_path.validator
    def validate_ssh_key_path(self, attribute, value):
        if value:
            Core.get_abs_path(value)


@attr.s
class Runner:

    executor = attr.ib(type=str)
    command = attr.ib(type=str)
    hosts: typing.List[Host] = attr.ib()

    @executor.validator
    def validate_type(self, attribute, value):
        if value not in ['powershell', 'cmd', 'ssh']:
            raise ValueError("Please provide a value for type of either 'powershell','cmd', or 'ssh'")



class State(object):
    """
    We define a state object which provides some utility functions for the
    individual states within the state machine.
    """

    def on_event(self, event):
        """
        Handle events that are delegated to this State.
        """
        pass

    def __repr__(self):
        """
        Leverages the __str__ method to describe the State.
        """
        return self.__str__()

    def __str__(self):
        """
        Returns the name of the State.
        """
        return self.__class__.__name__



import os
import logging.config
import yaml

from logging import FileHandler, DEBUG, INFO, ERROR, WARNING, CRITICAL
import logging


class DebugFileHandler(FileHandler):
    def __init__(self, filename, mode='a', encoding=None, delay=False):
        super().__init__(filename, mode, encoding, delay)

    def emit(self, record):
        if not record.levelno == DEBUG:
            return
        super().emit(record)


class LoggingBase(type):
    def __init__(cls, *args):
        super().__init__(*args)
        cls.setup_logging()

        # Explicit name mangling
        logger_attribute_name = '_' + cls.__name__ + '__logger'

        # Logger name derived accounting for inheritance for the bonus marks
        logger_name = '.'.join([c.__name__ for c in cls.mro()[-2::-1]])

        setattr(cls, logger_attribute_name, logging.getLogger(logger_name))

    def setup_logging(cls, default_path='./atomic_operator/data/logging.yml', default_level=logging.INFO, env_key='LOG_CFG'):
        """Setup logging configuration
        """
        path = os.path.abspath(os.path.expanduser(os.path.expandvars(default_path)))
        value = os.getenv(env_key, None)
        if value:
            path = value
        if os.path.exists(os.path.abspath(path)):
            with open(path, 'rt') as f:
                config = yaml.safe_load(f.read())
            logger = logging.config.dictConfig(config)
        else:
            logger = logging.basicConfig(level=default_level)



from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()

version = dict()
with open("./rudder/utils/version.py") as fp:
    exec(fp.read(), version)


setup(
    name='rudder',
    version=version['__version__'],
    packages=find_packages(exclude=['tests*']),
    license='MIT',
    description='A Python package to execute code remotely to multiple operating system platforms',
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    install_requires=required,
    keywords=['rudder', 'winrm', 'ssh', 'cmd', 'powershell'],
    url='https://github.com/MSAdministrator/rudder',
    author='MSAdministrator',
    author_email='rickardja@live.com',
    python_requires='>3.6',
    package_data={
        'rudder':  ['data/logging.yml']
    },
    entry_points={
          'console_scripts': [
              'rudder = rudder.__main__:main'
          ]
    }
)